/*******************************************************************************
 * Copyright (c) 2011 INRIA. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0 
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *     Wolfgang Kling - AtlanMod (EMN, INRIA, LINA)
 *     
 ******************************************************************************/

package org.eclipse.am3.moscript.atl.ext;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.gmt.am3.core.modeling.MetaElement;
import org.eclipse.gmt.am3.core.modeling.Model;
import org.eclipse.gmt.am3.core.modeling.ModelElement;
import org.eclipse.gmt.am3.extractor.launchers.utils.ATLUtils;
import org.eclipse.gmt.am3.modelhandler.emf.EMFModelElement;
import org.eclipse.gmt.am3.modelhandler.emf.EMFReferenceModel;
import org.eclipse.gmt.am3.platform.core.AM3Platform;
import org.eclipse.gmt.am3.platform.extension.globalmodelmanagement.GMMConstants;
import org.eclipse.gmt.am3.platform.extension.globalmodelmanagement.extensioninterfaces.ITransformationExecutor;
import org.eclipse.gmt.am3.platform.extension.globalmodelmanagement.transformationexecutionengine.TransformationExecutionHelper;
import org.eclipse.gmt.am3.platform.runtime.core.ModelUtil;
import org.eclipse.gmt.am3.platform.runtime.core.context.AM3Context;
import org.eclipse.gmt.am3.platform.runtime.core.impl.EMFModelUtil;
import org.eclipse.m2m.atl.core.ATLCoreException;
import org.eclipse.m2m.atl.core.IInjector;
import org.eclipse.m2m.atl.core.IModel;
import org.eclipse.m2m.atl.core.IReferenceModel;
import org.eclipse.m2m.atl.core.emf.EMFModel;
import org.eclipse.m2m.atl.core.emf.EMFModelFactory;
import org.eclipse.m2m.atl.core.service.CoreService;
import org.eclipse.m2m.atl.engine.emfvm.lib.ASMModule;
import org.eclipse.m2m.atl.engine.emfvm.lib.AbstractStackFrame;
import org.eclipse.m2m.atl.engine.emfvm.lib.ExecEnv;
import org.eclipse.m2m.atl.engine.emfvm.lib.LibExtension;
import org.eclipse.m2m.atl.engine.emfvm.lib.Operation;
import org.eclipse.m2m.atl.engine.emfvm.lib.Tuple;

/**
 * @author <a href="mailto:wolfgang.kling@inria.fr">Wolfgang Kling</a>
 */
public class MgMLibExtension implements LibExtension {

	private static Model am3Megamodel;
	public final static String refModelKey = "#MgM";
	public final static String modelKey = "#inMgM";

	/* (non-Javadoc)
	 * @see org.eclipse.m2m.atl.engine.emfvm.lib.LibExtension#apply(org.eclipse.m2m.atl.engine.emfvm.lib.ExecEnv, java.util.Map)
	 */
	@Override
	public void apply(ExecEnv execEnv, Map<String, Object> options) {

		am3Megamodel = AM3Platform.getCurrentMegaModel();

		IModel megamodel = execEnv.getModel(modelKey);
		if (megamodel != null) {
			EMFModel emfATLMegamodel = (EMFModel) megamodel;
			emfATLMegamodel
					.setResource(((org.eclipse.gmt.am3.modelhandler.emf.EMFModel) am3Megamodel)
							.getResource());
			emfATLMegamodel.setIsTarget(true);

			IReferenceModel refModel = megamodel.getReferenceModel();

			Object launchMetaElement = refModel
					.getMetaElementByName("ATLTransformationLaunch");
			if (launchMetaElement != null) {
				execEnv.registerOperation(launchMetaElement, opRun());
			} else {
				throw new Error(this.getClass().getCanonicalName()
						+ " - Unable to find ATLTransformationLaunch element");
			}

			Object m2mTransfMetaElement = refModel
					.getMetaElementByName(GMMConstants.GMM_M2MTRANSFORMATION);
			if (m2mTransfMetaElement != null) {
				execEnv.registerOperation(m2mTransfMetaElement, opApplyTo());
			} else {
				throw new Error(this.getClass().getCanonicalName()
						+ " - Unable to find "
						+ GMMConstants.GMM_M2MTRANSFORMATION + " element");
			}
			Object modelMetaElement = refModel.getMetaElementByName(GMMConstants.GMM_MODEL);
			if (modelMetaElement != null) {
				execEnv.registerOperation(modelMetaElement, opAllContentInstancesOf());
				execEnv.registerOperation(modelMetaElement, opSaveModel());
				execEnv.registerOperation(modelMetaElement, opInject());
			} 
			else {
				throw new Error(this.getClass().getCanonicalName()
						+ "Unable to find " + GMMConstants.GMM_MODEL
						+ " element");
			}

			Object locatedElement = refModel
					.getMetaElementByName("AM3Core::LocatedElement");
			if (locatedElement != null) {
				execEnv.registerOperation(locatedElement, opAvailable());
			} else {
				throw new Error(this.getClass().getCanonicalName()
						+ "Unable to find " + "AM3Core::LocatedElement"
						+ " element");
			}
			Object identifiedElement = refModel
					.getMetaElementByName("AM3Core::IdentifiedElement");
			if (identifiedElement != null) {
				execEnv.registerOperation(identifiedElement, opRemove());
				execEnv.registerOperation(identifiedElement, opUpdate());
			} else {
				throw new Error(this.getClass().getCanonicalName()
						+ "Unable to find " + "AM3Core::Element" + " element");
			}
			Object refModelMetaElement = refModel
					.getMetaElementByName(GMMConstants.GMM_REFERENCEMODEL);
			if (refModelMetaElement != null) {
				execEnv.registerOperation(refModelMetaElement,
						opRegisterInEMFpkgReg());
			} else {
				throw new Error(this.getClass().getCanonicalName()
						+ " - Unable to find "
						+ GMMConstants.GMM_REFERENCEMODEL + " element");
			}
			
			execEnv.registerOperation(ASMModule.class, opSaveModelIn());
			execEnv.registerOperation(ASMModule.class, opRegister());
			execEnv.registerOperation(ASMModule.class, opRegisterIn());
			execEnv.registerOperation(ASMModule.class,
					opCollectWrkSpaceFileNames());

		}
	}

	/**
	 * @param modelElement
	 * @return
	 */
	private static String getIdentifier(EObject modelElement) {
		EList<EReference> references = modelElement.eClass()
				.getEAllReferences();
		for (EReference ref : references) {
			if (ref.getName().equals(
					GMMConstants.AM3_FEATURE_IDENTIFIEDELEMENT_IDENTIFIER)) {
				EObject oRef = (EObject) modelElement.eGet(ref);
				EList<EAttribute> attributes = oRef.eClass()
						.getEAllAttributes();
				for (EAttribute attribute : attributes) {
					if (attribute.getName().equals(
							GMMConstants.AM3_FEATURE_IDENTIFIER_VALUE)) {
						return (String) oRef.eGet(attribute);
					}
				}
			}
		}
		return null;
	}

	/**
	 * @param megamodel
	 * @param id
	 * @param metaElementQualifiedName
	 * @return
	 */
	private static ModelElement getModelElement(Model megamodel, String id,
			String metaElementQualifiedName) {
		ModelUtil modelUtil = EMFModelUtil.getInstance();
		return modelUtil.getModelElementFromId(megamodel, id,
				metaElementQualifiedName);
	}

	/**
	 * @param model
	 * @param modelUri
	 * @return
	 * @throws Exception
	 */
	private static URI getRefModelURI(EObject model, URI modelUri) throws Exception {
		EList<EReference> references = model.eClass().getEAllReferences();
		for (EReference ref : references) {
			if (ref.getName().equals("conformsTo")) {
				EObject oRef = (EObject) model.eGet(ref);
				if (oRef != null) {
					return getTerminalModelURI(oRef);
				}
			}
		}

		throw new Exception("Cannot find ReferenceModel for Model "
				+ modelUri.toString());
	}

	/**
	 * @param model
	 * @return
	 * @throws Exception
	 */
	private static URI getTerminalModelURI(EObject model) throws Exception {
		EList<EReference> references = model.eClass().getEAllReferences();
		for (EReference ref : references) {
			if (ref.getName().equals("locator")) {
				EObject oRef = (EObject) model.eGet(ref);
				EList<EAttribute> attributes = oRef.eClass()
						.getEAllAttributes();
				for (EAttribute attribute : attributes) {
					if (attribute.getName().equals("value")) {
						return getURI((String) oRef.eGet(attribute));
					}
				}
			}
		}
		throw new Exception("Cannot find TerminalModel locator value");
	}

	/**
	 * @param locatorValue
	 * @return
	 */
	private static URI getURI(String locatorValue) {
		if (locatorValue.startsWith("/")) {// workspace relative path
			IPath workspcPath = Platform.getLocation();
			String path = workspcPath.toString() + locatorValue;
			return URI.createFileURI(path);
		} else {
			if (locatorValue.startsWith("uri:")) {
				locatorValue = locatorValue
						.substring(locatorValue.indexOf(":") + 1);
				return URI.createURI(locatorValue);
			} else if (locatorValue.startsWith("http:")
					|| locatorValue.startsWith("mem:")) {
				return URI.createURI(locatorValue);
			} else
				// local file system absolute path
				return URI.createFileURI(locatorValue);
		}
	}

	/**
	 * @return
	 */
	private static Operation opAllContentInstancesOf() {
		return new Operation(2, "allContentInstancesOf") {

			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();

				MgMEnv mgmEnv = MgMEnv.getInstance();
				Set<Object> ret = new LinkedHashSet<Object>();
				try {

					URI modelUri = getTerminalModelURI((EObject) localVars[0]);
					URI refModelUri = getRefModelURI((EObject) localVars[0], modelUri);

					IReferenceModel referenceModel = (IReferenceModel) (frame
							.getExecEnv().getModelsByName()).get(refModelUri
							.toString());
					if (referenceModel == null)
						referenceModel = mgmEnv.loadReferenceModel(refModelUri);

					IModel model = null;
					String modelUriStr = modelUri.toString();
					model = (frame.getExecEnv().getModelsByName()).get(modelUriStr);

					if (model == null)
						model = mgmEnv.loadModel(modelUri, referenceModel);

					String type = (String) localVars[1];
					EObject typeEO = (EObject) model.getReferenceModel()
							.getMetaElementByName(type);

					if (typeEO != null) {
						frame.getExecEnv().addModel(refModelUri.toString(),
								referenceModel);
						frame.getExecEnv().addModel(modelUri.toString(), model);
						ret.addAll(model.getElementsByType(typeEO));
					}

					return ret;
				} catch (Exception e) {
					e.printStackTrace();
					return ret;
				}
			}
		};
	}

	/**
	 * @return
	 */
	private static Operation opApplyTo() {
		return new Operation(2, "applyTo") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject transf = (EObject) localVars[0];
				HashMap<String, EObject> transfParams = (HashMap<String, EObject>) localVars[1];

				String identifier = MgMLibExtension.getIdentifier(transf);
				ModelElement m2mTransformation = getModelElement(am3Megamodel,
						identifier, GMMConstants.GMM_M2MTRANSFORMATION);

				String transformationType = m2mTransformation.getMetaElement()
						.getFullyQualifiedName();
				ITransformationExecutor executor = TransformationExecutionHelper
						.getInstance().getExecutorFor(transformationType);
				if (executor != null
						&& transformationType
								.equals("GMM4ATL::ATLTransformation")) {

					Map<ModelElement, ModelElement> models = new HashMap();

					ArrayList<ModelElement> transformationParameters = (ArrayList<ModelElement>) m2mTransformation
							.get(GMMConstants.GMM_FEATURE_TRANSFORMATION_PARAMETERS);
					for (ModelElement transformationParam : transformationParameters) {
						ModelElement moduleParameter = (ModelElement) transformationParam
								.get("moduleParameter");
						// This is particular to ATL implementation,
						// formalModelName should generalized to all m2m
						// transformations
						String formalModelName = (String) moduleParameter
								.get("formalModelName");
						EObject stackModelElement = transfParams
								.get(formalModelName);

						ModelElement modelModelElem = null;
						if (stackModelElement != null) {
							String id = MgMLibExtension
									.getIdentifier(stackModelElement);
							modelModelElem = getModelElement(am3Megamodel, id,
									GMMConstants.GMM_MODEL);
						}
						models.put(transformationParam, modelModelElem);
					}

					Map<String, Object> trResult = executor
							.executeNoSideEfects(m2mTransformation, models,
									new HashMap<String, Object>());

					if(trResult == null) {
						throw new Error("Transformation " + identifier + " finished with errors");
					}
					ModelElement transformationRecord = (ModelElement) trResult
							.get("trRecord");
					Map<String, IModel> physicalModels = (Map<String, IModel>) trResult
							.get("models");

					// Puts the newly created model elements into the output
					// models map that will be returned to ATL
					HashMap<String, EObject> outputModels = new HashMap<String, EObject>();
					ArrayList<ModelElement> transfParamValues = (ArrayList<ModelElement>) transformationRecord
							.get("values");
					for (ModelElement transfParamValue : transfParamValues) {
						ModelElement transfParameter = (ModelElement) transfParamValue
								.get("type");
						ModelElement outputmodelElement = (ModelElement) transfParamValue
								.get("value");
						EObject outputModel = ((EMFModelElement) outputmodelElement)
								.getEObject();

						if (transfParameter.get("direction").toString().equals(
								GMMConstants.GMM_ENUM_DIRECTION_OUT)) {
							ModelElement atlParameter = (ModelElement) transfParameter
									.get("moduleParameter");
							String formalModelName = (String) atlParameter
									.get("formalModelName");
							// Loads the newly created models (IModel) to the
							// ATL execEnv
							ModelElement outputmodelLocator = (ModelElement) outputmodelElement
									.get(GMMConstants.AM3_FEATURE_LOCATEDELEMENT_LOCATOR);
							String outputmodelStrLocator = (String) outputmodelLocator
									.get(GMMConstants.AM3_FEATURE_LOCATOR_VALUE);
							outputModels.put(formalModelName, outputModel);
							frame.getExecEnv().addModel(outputmodelStrLocator,
									physicalModels.get(formalModelName));
						}
					}

					return outputModels;
				} else {
					throw new Error("Could not find appropiate executor for "
							+ transformationType + " transformation type");
				}
			}

		};
	}

	//TODO This operation is using the injectors registered in ATL, it should use instead
	//the injectors registered in the Megamodel and thus in the AM3 platform.
	//The megamodel should distinguish between M2M, T2M and M2T taking also in account themplate
	//based aproaches like acceleo, xpand etc.
	/**
	 * @return
	 */
	private static Operation opInject() {
		return new Operation(3, "inject") {

			@Override
			public Object exec(AbstractStackFrame frame) {

				Object[] localVars = frame.getLocalVars();
				String injector = (String) localVars[1];

				try {
					String injectors[] = CoreService.getInjectorsNames();
					for (int i = 0; i < injectors.length; i++) {
						if (injectors[i].equalsIgnoreCase(injector)) {
							
							URI modelUri = getTerminalModelURI((EObject) localVars[0]);
							URI refModelUri = getRefModelURI((EObject) localVars[0], modelUri);
							
							
							IReferenceModel referenceModel = (IReferenceModel) 
								(frame.getExecEnv().getModelsByName()).get(refModelUri.toString());
							
							if (referenceModel == null){
								MgMEnv mgmEnv = MgMEnv.getInstance();
								referenceModel = mgmEnv.loadReferenceModel(refModelUri);
							}
							
							IInjector injectorInstance = CoreService.getInjector(injector);
							
							EMFModelFactory atl3modelFactory = (EMFModelFactory) CoreService.getModelFactory("EMF");
							IModel model = atl3modelFactory.newModel(referenceModel);
							injectorInstance.inject(model, modelUri.toString().replace("file:/", "file://"));
							

							Tuple parameters = (Tuple) localVars[2];
							Map locatedElement = parameters.getMap();
							HashSet value = (HashSet) locatedElement.get("value");
							Iterator it = value.iterator();
							while(it.hasNext()) {
								Tuple attribute = (Tuple)it.next();
								if(((String)attribute.getMap().get("name")).equals("locator")) {
									Map locatorMap = attribute.getMap();
									Iterator it2 = ((HashSet) locatorMap.get("value")).iterator();
									while(it2.hasNext()) {
										Tuple locatorValue = (Tuple)it2.next();
										String locStr = ((String)locatorValue.getMap().get("value"));
										locatorValue.getMap().put("value", "mem:/" + locStr);
										ModelElement terminalModel = MgMLibExtension.registerElement(parameters);
										EObject outputModel = ((EMFModelElement) terminalModel).getEObject();
										String strUri = getTerminalModelURI((EObject) localVars[0]).toString();
										frame.getExecEnv().addModel("mem:/" + locStr, model);
										System.out.println(strUri);
										return outputModel;
									}
								}
							}
							
						}
					}

				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "No injector with name " + injector
						+ " registered in the ATLVM for use";
			}
		};

		// The following is the correct way to implement the injector but it
		// relies on GMM4TCS and
		// ModelToTextTransformations should not be abstract so it can be used
		// generically.
		// Also the injector should be associated to the concreteTextualSyntax
		// somehow in the Megamodel
		/*
		 * return new Operation(2, "inject") {
		 * 
		 * @Override public Object exec(AbstractStackFrame frame) { Object[]
		 * localVars = frame.getLocalVars(); EObject ext2modelTransfEObjt =
		 * (EObject) localVars[0]; EObject entity = (EObject) localVars[1];
		 * 
		 * String identifier =
		 * MgMLibExtension.getIdentifier(ext2modelTransfEObjt); ModelElement
		 * ext2modelTransf = MgMLibExtension.getModelElement(am3Megamodel,
		 * identifier, GMMConstants.GMM_E2MTRANSFORMATION);
		 * 
		 * String transformationType =
		 * ext2modelTransf.getMetaElement().getFullyQualifiedName();
		 * ITransformationExecutor executor =
		 * TransformationExecutionHelper.getInstance
		 * ().getExecutorFor(transformationType); if(executor != null) {
		 * 
		 * Map<ModelElement, ModelElement> arguments = new HashMap<ModelElement,
		 * ModelElement>();
		 * 
		 * String entityId = MgMLibExtension.getIdentifier(entity); ModelElement
		 * entityModelElem = getModelElement(am3Megamodel, entityId,
		 * "GlobalModelManagement::Entity"); ArrayList<ModelElement>
		 * externalParameters = (ArrayList<ModelElement>)
		 * ext2modelTransf.get("externalParameters"); //The external parameter
		 * is suposed to be the grammar but can be null in this case because all
		 * what is needed is in the jar //generated by TCS
		 * arguments.put(externalParameters.get(0), entityModelElem);
		 * 
		 * ArrayList<ModelElement> modelParameters = (ArrayList<ModelElement>)
		 * ext2modelTransf
		 * .get(GMMConstants.GMM_FEATURE_MODELTRANSFORMATION_MODELPARAMETERS);
		 * //It is not necesary to put a ModelElement as value of the map if we
		 * don't want to serialize it directly
		 * arguments.put(modelParameters.get(0), modelParameters.get(0));
		 * 
		 * Map<String, Object> trResult =
		 * executor.executeNoSideEfects(ext2modelTransf, arguments, null);
		 * IModel physicalModel = (IModel) trResult.get("model");
		 * 
		 * ModelElement transformationRecord = (ModelElement)
		 * trResult.get("trRecord"); ArrayList<ModelElement> transfParamValues =
		 * (ArrayList<ModelElement>)transformationRecord.get("values");
		 * ModelElement transfParamValue = transfParamValues.get(0);
		 * ModelElement outputmodelElement = (ModelElement)
		 * transfParamValue.get("value"); EObject outputModel =
		 * ((EMFModelElement)outputmodelElement).getEObject(); //Loads the newly
		 * created models (IModel) to the ATL execEnv ModelElement
		 * outputmodelLocator = (ModelElement)
		 * outputmodelElement.get(GMMConstants
		 * .AM3_FEATURE_LOCATEDELEMENT_LOCATOR); String outputmodelStrLocator =
		 * (String)
		 * outputmodelLocator.get(GMMConstants.AM3_FEATURE_LOCATOR_VALUE);
		 * frame.getExecEnv().addModel(outputmodelStrLocator, physicalModel);
		 * 
		 * return outputModel; } else { throw new
		 * Error("Could not find appropiate executor for " + transformationType
		 * + " transformation type"); } } };
		 */
	}

	//TODO This operation should not be implemented here, it should
	// be implemented in the abstract and concrete syntax and the compiler.
	/**
	 * @return
	 */
	private static Operation opRegister() {
		return new Operation(2, "register") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				ASMModule asmModule = (ASMModule) localVars[0];
				Tuple parameters = (Tuple) localVars[1];
				MgMLibExtension.registerElement(parameters);
				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}

	//TODO This operation should not be implemented here, it should
	// be implemented in the abstract and concrete syntax and the compiler.
	/**
	 * @return
	 */
	private static Operation opRemove() {
		return new Operation(2, "remove") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject elementEObject = (EObject) localVars[0];
				String identifier = MgMLibExtension
						.getIdentifier(elementEObject);
				ModelElement identifiedElement = getModelElement(am3Megamodel,
						identifier, "AM3Core::IdentifiedElement");
				am3Megamodel.remove(identifiedElement);
				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}

	/**
	 * @return
	 */
	private static Operation opRun() {
		return new Operation(2, "run") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject launch = (EObject) localVars[0];
				//TODO Implement the operation, or it can be also be done through 
				//MoScript libraries.
				return "operation has not been yet implemented";
			}
		};
	}

	//TODO This operation should not be implemented here, it should
	// be implemented in the abstract and concrete syntax and the compiler.
	/**
	 * @return
	 */
	private static Operation opSaveModel() {
		return new Operation(2, "save") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				String locator = null;
				try {
					locator = (String) localVars[1];
					
					URI modelUri = getTerminalModelURI((EObject) localVars[0]);
					IModel model = (frame.getExecEnv().getModelsByName())
							.get(modelUri.toString());
					ATLUtils.getInstance().getExtractor().extract(model, locator);
					
					//ModelElement newLocator = AM3Platform.getCurrentMegaModel().createModelElement(
					//		AM3Platform.getAM3Metamodel().getMetaElementByName(GMMConstants.AM3_EFSLOCATOR)
					//		);
					//newLocator.set(GMMConstants.AM3_FEATURE_LOCATOR_VALUE, locator);
					
					EObject locatedElementEObject = (EObject) localVars[0];
					String identifier = MgMLibExtension.getIdentifier(locatedElementEObject);
					//udpdate locator in metadata to point to the filesystem
					ModelElement locatedElement = getModelElement(am3Megamodel, identifier, "AM3Core::LocatedElement");
					ModelElement modelLocator = (ModelElement) locatedElement.get(GMMConstants.AM3_FEATURE_LOCATEDELEMENT_LOCATOR);
					modelLocator.set(GMMConstants.AM3_FEATURE_LOCATOR_VALUE, locator);
				    //locatedElement.set(GMMConstants.AM3_FEATURE_LOCATEDELEMENT_LOCATOR, newLocator);

				} catch (Exception e) {
					e.printStackTrace();
				}
				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}

	//TODO This operation should not be implemented here, it should
	// be implemented in the abstract and concrete syntax and the compiler.
	/**
	 * @return
	 */
	private static Operation opSaveModelIn() {
		return new Operation(5, "_saveModelIn") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				ASMModule asmModule = (ASMModule) localVars[0];
				EObject model = (EObject) localVars[1];
				String locator = (String) localVars[2];
				String identifier = (String) localVars[3];
				EObject trgtMegaModel = (EObject) localVars[4];
				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}

	//TODO This operation should not be implemented here, it should
	// be implemented in the abstract and concrete syntax and the compiler.
	/**
	 * @return
	 */
	private static Operation opUpdate() {
		return new Operation(2, "update") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject locatedElementEObject = (EObject) localVars[0];
				Tuple parameters = (Tuple) localVars[1];
				String identifier = MgMLibExtension
						.getIdentifier(locatedElementEObject);
				ModelElement locatedElement = getModelElement(am3Megamodel,
						identifier, "AM3Core::LocatedElement");
				am3Megamodel.remove(locatedElement);

				MgMLibExtension.registerElement(parameters);
				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}

	/**
	 * @param modelElement
	 * @param metaElement
	 * @param elementContent
	 * @param referencedElementsCache 
	 * @param referencingElementsCache 
	 */
	private static void populateElement(ModelElement modelElement,
			MetaElement metaElement, Set elementContent, Map referencedElementsCache, ArrayList<Object[]> referencingElementsCache) {
		Iterator it = elementContent.iterator();
		while (it.hasNext()) {
			Object ob = it.next();
			if (ob instanceof Tuple) {
				Tuple attribute = (Tuple) ob;
				//System.out.println("************** " + attribute);
				String name = (String) attribute.get(null, "name");
				try {
					metaElement.getFeature(name);
				} catch (Throwable ex) {
					System.out.println("Feature name " + name
							+ " not found in metamodel \\n" + attribute.toString());
					throw new Error(ex);
				}
				if (metaElement.getFeature(name).isPrimitive()) {
					Object value = attribute.get(null, "value");
					modelElement.set(name, value);
				} else {
					String concreteType = (String) attribute.get(null,
							"concreteType");
					Object value = attribute.get(null, "value");
					if (value instanceof String && ((String)value).startsWith("@")) {
						String crossRefKey = ((String)value);
						Object[] elemNattrNRef = new Object[] {modelElement, name, crossRefKey};  // legal
						referencingElementsCache.add(elemNattrNRef);
					} else if (value instanceof EObject) {
						modelElement.set(name, attribute.get(null, "value"));
					} else if (metaElement.getFeature(name).isMultiValued()) {
						Collection values = (Collection) value;
						Iterator it2 = values.iterator();
						while (it2.hasNext()) {
							MetaElement newMetaElement = am3Megamodel
									.getReferenceModel().getMetaElementByName(
											concreteType);
							ModelElement newModelElement = am3Megamodel
									.createModelElement(newMetaElement);
							Object content = it2.next();
							Map tupleMap = ((Tuple)content).getMap();
							String creationId = (String) tupleMap.get("creationId");
							if(creationId != null)
								referencedElementsCache.put(creationId, newModelElement);
	
							populateElement(newModelElement, newMetaElement,
									(Set)tupleMap.get("value"), referencedElementsCache, referencingElementsCache);
							modelElement.add(name, newModelElement);
						}
					} else {
						MetaElement newMetaElement = am3Megamodel
								.getReferenceModel().getMetaElementByName(
										concreteType);
						ModelElement newModelElement = am3Megamodel
								.createModelElement(newMetaElement);

						String creationId = (String) attribute.get(null, "creationId");
						if(creationId != null)
							referencedElementsCache.put(creationId, newModelElement);

						populateElement(newModelElement, newMetaElement,
								(Set) attribute.get(null, "value"), referencedElementsCache, referencingElementsCache);
						modelElement.set(name, newModelElement);
					}
				}

			}
		}
	}

	/**
	 * @param element
	 * @return
	 */
	public static ModelElement registerElement(Tuple element) {
		Map referencedElementsCache = new HashMap();
		ArrayList<Object[]> referencingElementsCache = new ArrayList<Object[]>();
		String concreteType = (String) element.get(null, "concreteType");
		Set elementContent = (Set) element.get(null, "value");

		MetaElement newMetaElement = am3Megamodel.getReferenceModel()
				.getMetaElementByName(concreteType);
		ModelElement newModelElement = am3Megamodel
				.createModelElement(newMetaElement);
		String creationId = (String) element.get(null, "creationId");

		if(creationId != null) 
			referencedElementsCache.put(creationId, newModelElement);

		populateElement(newModelElement, newMetaElement, elementContent, referencedElementsCache, referencingElementsCache);
		resolveCrossRefs(referencingElementsCache, referencedElementsCache);
		return newModelElement;
	}

	private static void resolveCrossRefs(ArrayList<Object[]> referencingElementsCache, Map referencedElementsCache) {

		for(Object[] referencingElement : referencingElementsCache) {
			ModelElement modelElement = (ModelElement) referencingElement[0];
			String attributeName = (String) referencingElement[1];
			String crossRefKey = (String) referencingElement[2];
			if(referencedElementsCache.get(crossRefKey) == null)
				System.out.println("Error, cross reference " + crossRefKey + "could not be resolved " + "check if element has a creation id");
			else
				modelElement.set(attributeName, referencedElementsCache.get(crossRefKey));
		}
	}

	/**
	 * @param model
	 * @param uri
	 * @throws Exception
	 */
	private static void save(IModel model, URI uri) throws Exception {

		if (model.isTarget()) {
			ATLUtils.getInstance().getExtractor()
					.extract(model, uri.toString());
		}
	}

	/**
	 * @return
	 */
	private Operation opAvailable() {
		return new Operation(1, "available") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject locatedElementEObject = (EObject) localVars[0];
				String identifier = MgMLibExtension
						.getIdentifier(locatedElementEObject);
				ModelElement locatedElement = getModelElement(am3Megamodel,
						identifier, "AM3Core::LocatedElement");
				ModelElement locator = (ModelElement) locatedElement
						.get("locator");
				String locatorValue = (String) locator.get("value");
				URI uri = getURI(locatorValue);
				if (uri.toString().startsWith("file:"))
					return new File(uri.toString().replaceFirst("file:", ""))
							.exists();
				else if (uri.toString().startsWith("http:")) {
					ResourceSet resSet = new ResourceSetImpl();
					try {
						if (resSet.getResource(uri, true) != null) {
							return true;
						} else {
							return false;
						}
					} catch (Exception e) {
						// e.printStackTrace();
						return false;
					}

				} else {
					return false;
				}
			}
		};
	}

	//TODO this operation should be asigned to an element of the megamodel instead to
	//thisModule, for example a Megamodel element called Wokspace or System.
	/**
	 * @return
	 */
	private Operation opCollectWrkSpaceFileNames() {
		return new Operation(1, "collectWrkSpaceFileNames") {

			@Override
			public Object exec(AbstractStackFrame frame) {

				ArrayList<String> paths = new ArrayList<String>();
				IProject[] projects = ResourcesPlugin.getWorkspace().getRoot()
						.getProjects();
				for (IProject project : projects) {
					try {
						if (project.isOpen()) {
							IResource[] resources = project.members();
							for (IResource resource : resources) {
								if (resource instanceof IFile) {
									paths.add(resource.getFullPath()
											.toOSString());
								} else {
									getInnerFiles((IFolder) resource, paths);
								}
							}
						}
					} catch (CoreException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
				return paths;
			}

			private void getInnerFiles(IFolder folder, ArrayList<String> trgt) {
				try {
					IResource[] resources = folder.members();
					for (IResource resource : resources) {
						if (resource instanceof IFolder) {
							getInnerFiles((IFolder) resource, trgt);
						} else {
							trgt.add(resource.getFullPath().toOSString());
						}
					}

				} catch (CoreException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		};
	}

	//TODO This operation should not be implemented here, it should
	// be implemented in the abstract and concrete syntax and the compiler.
	/**
	 * @return
	 */
	private Operation opRegisterIn() {
		return new Operation(4, "registerIn") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				ASMModule asmModule = (ASMModule) localVars[0];
				EObject model = (EObject) localVars[1];
				String locator = (String) localVars[2];
				String identifier = (String) localVars[3];
				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}

	/**
	 * @return
	 */
	private Operation opRegisterInEMFpkgReg() {
		return new Operation(1, "registerInEMFpkgReg") {

			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject refModel = (EObject) localVars[0];
				String identifier = MgMLibExtension.getIdentifier(refModel);
				ModelElement refModelElement = getModelElement(am3Megamodel,
						identifier, GMMConstants.GMM_REFERENCEMODEL);
				ModelElement locator = (ModelElement) refModelElement
						.get(GMMConstants.AM3_FEATURE_LOCATEDELEMENT_LOCATOR);
				String locatorValue = (String) locator
						.get(GMMConstants.AM3_FEATURE_LOCATOR_VALUE);
				URI uri = getURI(locatorValue);
				MgMEnv mgmEnv = MgMEnv.getInstance();
				try {
					IReferenceModel referenceModel = mgmEnv
							.loadReferenceModel(uri);
					org.eclipse.m2m.atl.core.emf.EMFReferenceModel atlEmfReferenceModel = (org.eclipse.m2m.atl.core.emf.EMFReferenceModel) referenceModel;
					EMFReferenceModel am3EmfReferenceModel = new EMFReferenceModel(
							atlEmfReferenceModel.getResource());
					AM3Context.getDefault().getCoreRuntime().addToAM3(uri,
							am3EmfReferenceModel);
				} catch (ATLCoreException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				//TODO instead of returning a message it should throw an exception if
				//it is not possible to conclude the operation.
				return "ignore this, this operation should not return anything";
			}
		};
	}
}
