/*******************************************************************************
 * Copyright (c) 2011 INRIA, Vrije Universiteit Brussel. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0 
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Original Contributors: 
 *    Frederic Jouault (INRIA) - initial API and implementation
 *    Dennis Wagelaar (Vrije Universiteit Brussel)
 * Adaptation to MoScript:
 *    Wolfgang Kling - AtlanMod (EMN, INRIA, LINA)
 *  
 ******************************************************************************/
package org.eclipse.am3.moscript.engine.compiler;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.ecore.EObject;

/**
 * @author <a href="mailto:frederic.jouault@univ-nantes.fr">Frederic Jouault</a>
 * @author <a href="mailto:dennis.wagelaar@vub.ac.be">Dennis Wagelaar</a>
 * @author <a href="mailto:wolfgang.kling@inria.fr">Wolfgang Kling</a>
 */

/**
 * The MoScript compiler.
 */
public final class Compiler {

	private static final int MAX_LINE_LENGTH = 1000;

	private Compiler() {
		super();
	}

	/**
	 * Searches for the correct implementation of {@link StandaloneCompiler} which is independent of
	 * Eclipse platform stuff like IFile, IResource, etc.
	 * 
	 * @param compilerName
	 *            name of the compiler to search for
	 * @return the compiler which was found
	 */
	public static StandaloneCompiler getCompiler(String compilerName) {
		return new DefaultCompiler();
	}

	/**
	 * Performs compilation.
	 * 
	 * @param in
	 *            The InputStream to get MoScript source from.
	 * @param out
	 *            The IFile to which the MoScript compiled program will be saved.
	 * @return the problems which occured during compilation
	 */
	public static EObject[] compile(InputStream in, IFile out) throws IOException {
		EObject[] ret = compile(in, out.getLocation().toString());
		try {
			out.refreshLocal(0, null);
		} 
		catch (CoreException e) {
			e.printStackTrace();
		}
		return ret;
	}

	/**
	 * Standalone compilation.
	 * 
	 * @param in
	 *            The InputStream to get MoScript source from.
	 * @param outputFileName
	 *            The output file name
	 * @return the problems which occurred during compilation
	 */
	public static EObject[] compile(InputStream in, String outputFileName) throws IOException {
		EObject[] ret = null;
		InputStream newIn = in;
		// The BufferedInputStream is required to reset the stream before actually compiling
		newIn = new BufferedInputStream(newIn, MAX_LINE_LENGTH);
		newIn.mark(MAX_LINE_LENGTH);
		byte[] buffer = new byte[MAX_LINE_LENGTH];
		newIn.read(buffer);
		newIn.reset();

		ret = getCompiler("default").compileWithProblemModel(newIn, outputFileName);
		return ret;
	}

	/**
	 * Standalone compilation, to an outputStream.
	 * 
	 * @param in
	 *            The InputStream to get MoScript source from.
	 * @param outputStream
	 *            The output file name
	 * @return the problems which occurred during compilation
	 */
	public static EObject[] compile(InputStream in, OutputStream outputStream) throws IOException {
		EObject[] ret = null;
		InputStream newIn = in;
		// The BufferedInputStream is required to reset the stream before actually compiling
		newIn = new BufferedInputStream(newIn, MAX_LINE_LENGTH);
		newIn.mark(MAX_LINE_LENGTH);
		byte[] buffer = new byte[MAX_LINE_LENGTH];
		newIn.read(buffer);
		newIn.reset();

		ret = getCompiler("default").compileWithProblemModel(newIn, outputStream);
		return ret;
	}

}
